/* template.c
 * Linux CAN-bus device driver.
 * Written by Arnaud Westenberg email:arnaud@casema.net
 * This software is released under the GPL-License.
 * Version 0.6  18 Sept 2000
 */ 

/* This file is intended as a template file for currently unsupported hardware.
 * Once you've changed/added the functions specific to your hardware it is
 * possible to load the driver with the hardware option hw=template.
 */

#include <linux/autoconf.h>
#if defined (CONFIG_MODVERSIONS) && !defined (MODVERSIONS)
#define MODVERSIONS
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/ioport.h>
#include <linux/delay.h>
#include <asm/errno.h>
#include <asm/io.h>

#include "../include/main.h"
#include "../include/template.h"
#include "../include/i82527.h"
#include "../include/sja1000.h"

unsigned adsgm_read_register(unsigned long address);
void adsgm_write_register(unsigned char data, unsigned long address);

/*
 * IO_RANGE is the io-memory range that gets reserved, please adjust according
 * your hardware. Example: #define IO_RANGE 0x100 for i82527 chips or
 * #define IO_RANGE 0x20 for sja1000 chips in basic CAN mode.
 */
#define IO_RANGE 0x400
#define IO_RAM   0x10
#define HIGH

static unsigned int base_io  = 0xf1000000;


/**
 * adsgm_request_io: - reserve io memory
 * @io_addr: The reserved memory starts at @io_addr, wich is the module 
 * parameter @io.
 *
 * The function adsgm_request_io() is used to reserve the io-memory. If your
 * hardware uses a dedicated memory range as hardware control registers you
 * will have to add the code to reserve this memory as well. 
 * %IO_RANGE is the io-memory range that gets reserved, please adjust according
 * your hardware. Example: #define IO_RANGE 0x100 for i82527 chips or
 * #define IO_RANGE 0x20 for sja1000 chips in basic CAN mode.
 * Return Value: The function returns zero on success or %-ENODEV on failure
 * File: src/template.c
 */
int adsgm_request_io(unsigned long io_addr)
{
	if (check_region(io_addr,IO_RANGE)) {
		CANMSG("Unable to open port: 0x%lx\n",io_addr);
		return -ENODEV;
	}
	else {
		request_region(io_addr,IO_RANGE,DEVICE_NAME);
		DEBUGMSG("Registered IO-memory: 0x%lx - 0x%lx\n", io_addr, io_addr + IO_RANGE - 1);
	}
    base_io = io_addr;
	return 0;
}

/**
 * adsgm_release_io - free reserved io-memory
 * @io_addr: Start of the memory range to be released.
 *
 * The function adsgm_release_io() is used to free reserved io-memory.
 * In case you have reserved more io memory, don't forget to free it here.
 * IO_RANGE is the io-memory range that gets released, please adjust according
 * your hardware. Example: #define IO_RANGE 0x100 for i82527 chips or
 * #define IO_RANGE 0x20 for sja1000 chips in basic CAN mode.
 * Return Value: The function always returns zero
 * File: src/template.c
 */
int adsgm_release_io(unsigned long io_addr)
{
	release_region(io_addr,IO_RANGE);

	return 0;
}

/**
 * adsgm_reset - hardware reset routine
 * @card: Number of the hardware card.
 *
 * The function adsgm_reset() is used to give a hardware reset. This is 
 * rather hardware specific so I haven't included example code. Don't forget to 
 * check the reset status of the chip before returning.
 * Return Value: The function returns zero on success or %-ENODEV on failure
 * File: src/template.c
 */
int adsgm_reset(int card) {
	int i=0;
	unsigned long saved_MSC2;

	DEBUGMSG("Resetting %s hardware ...\n", candevices_p[card]->hwname);

	// 8/14/01 Woojung
	// This is for GC+.
	// i82527 is connected on nCS3 of SA1110 on GC+.
	// It uses variable latency I/O.
	// But, if there is no physical CAN chip on that address,
	//  system(SA1110) may hang when acessing memory.
	// 
	// To avoid this problem, during reset and check H/W, 
	//  reconfigure MSC2 not to use variable latency I/O
	saved_MSC2 = MSC2;
	MSC2 &= 0xFFFFFFFC;
	// 8/24/01 Woojung
	// give some delay to make stable
	mdelay(5);

	/* Check hardware reset status */
	i=0;
	while ( (adsgm_read_register(base_io+iCPU) & iCPU_RST) && (i<=15)) {
		udelay(20000);
		i++;
	}
	if (i>=15) {
		CANMSG("Reset status timeout!\n");
		CANMSG("Please check your hardware.\n");
		// Restore MSC2
		MSC2 = saved_MSC2;
		return -ENODEV;
	}
	else {
		adsgm_write_register(0x07, base_io+iSTAT);
		if ((adsgm_read_register(base_io+iSTAT)&0x0F) != 0x07) {
			CANMSG("No i82527 Found!\n");
			CANMSG("Please check your hardware.\n");
			// Restore MSC2
			MSC2 = saved_MSC2;
			return -ENODEV;
		}
		else {
			DEBUGMSG("Chip0 reset status ok.\n");
		}
	}
	
	// Restore MSC2
	MSC2 = saved_MSC2;
	// 8/24/01 Woojung
	// give some delay to make stable
	mdelay(5);
	return 0;
}

#define RESET_ADDR 0x0
#define NR_82527 0
#define NR_SJA1000 0

/**
 * adsgm_init_hw_data - Initialze hardware cards
 * @card: Number of the hardware card.
 *
 * The function adsgm_init_hw_data() is used to initialize the hardware
 * structure containing information about the installed CAN-board.
 * %RESET_ADDR represents the io-address of the hardware reset register.
 * %NR_82527 represents the number of intel 82527 chips on the board.
 * %NR_SJA1000 represents the number of philips sja1000 chips on the board.
 * The flags entry can currently only be %PROGRAMMABLE_IRQ to indicate that
 * the hardware uses programmable interrupts.
 * Return Value: The function always returns zero
 * File: src/template.c
 */
int adsgm_init_hw_data(int card) 
{
	candevices_p[card]->res_addr=RESET_ADDR;
	candevices_p[card]->nr_82527_chips=1;
	candevices_p[card]->nr_sja1000_chips=0;
	candevices_p[card]->flags = 0; /*PROGRAMMABLE_IRQ;*/

	return 0;
}

#define CHIP_TYPE "i82527"
/**
 * adsgm_init_chip_data - Initialize chips
 * @card: Number of the hardware card
 * @chipnr: Number of the CAN chip on the hardware card
 *
 * The function adsgm_init_chip_data() is used to initialize the hardware
 * structure containing information about the CAN chips.
 * %CHIP_TYPE represents the type of CAN chip. %CHIP_TYPE can be "i82527" or
 * "sja1000".
 * The @chip_base_addr entry represents the start of the 'official' memory map
 * of the installed chip. It's likely that this is the same as the @io_addr
 * argument supplied at module loading time.
 * The @clock entry holds the chip clock value in Hz.
 * The entry @sja_cdr_reg holds hardware specific options for the Clock Divider
 * register. Options defined in the %sja1000.h file:
 * %CDR_CLKOUT_MASK, %CDR_CLK_OFF, %CDR_RXINPEN, %CDR_CBP, %CDR_PELICAN
 * The entry @sja_ocr_reg holds hardware specific options for the Output Control
 * register. Options defined in the %sja1000.h file:
 * %OCR_MODE_BIPHASE, %OCR_MODE_TEST, %OCR_MODE_NORMAL, %OCR_MODE_CLOCK,
 * %OCR_TX0_LH, %OCR_TX1_ZZ.
 * The entry @int_clk_reg holds hardware specific options for the Clock Out
 * register. Options defined in the %i82527.h file:
 * %iCLK_CD0, %iCLK_CD1, %iCLK_CD2, %iCLK_CD3, %iCLK_SL0, %iCLK_SL1.
 * The entry @int_bus_reg holds hardware specific options for the Bus 
 * Configuration register. Options defined in the %i82527.h file:
 * %iBUS_DR0, %iBUS_DR1, %iBUS_DT1, %iBUS_POL, %iBUS_CBY.
 * The entry @int_cpu_reg holds hardware specific options for the cpu interface
 * register. Options defined in the %i82527.h file:
 * %iCPU_CEN, %iCPU_MUX, %iCPU_SLP, %iCPU_PWD, %iCPU_DMC, %iCPU_DSC, %iCPU_RST.
 * Return Value: The function always returns zero
 * File: src/template.c
 */
int adsgm_init_chip_data(int card, int chipnr)
{
	candevices_p[card]->chip[chipnr]->chip_type=CHIP_TYPE;
	candevices_p[card]->chip[chipnr]->chip_base_addr=candevices_p[card]->io_addr;
	candevices_p[card]->chip[chipnr]->clock = 16000000;
    candevices_p[card]->chip[chipnr]->int_cpu_reg = iCPU_DSC; /*| iCPU_MUX; */
	candevices_p[card]->chip[chipnr]->int_clk_reg = iCLK_SL1;
	candevices_p[card]->chip[chipnr]->int_bus_reg = iBUS_CBY; /* | iBUS_DR1; */

	return 0;
}

/**
 * adsgm_init_obj_data - Initialize message buffers
 * @chipnr: Number of the CAN chip
 * @objnr: Number of the message buffer
 *
 * The function adsgm_init_obj_data() is used to initialize the hardware
 * structure containing information about the different message objects on the
 * CAN chip. In case of the sja1000 there's only one message object but on the
 * i82527 chip there are 15.
 * The code below is for a i82527 chip and initializes the object base addresses
 * The entry @obj_base_addr represents the first memory address of the message 
 * object. In case of the sja1000 @obj_base_addr is taken the same as the chips
 * base address.
 * Unless the hardware uses a segmented memory map, flags can be set zero.
 * Return Value: The function always returns zero
 * File: src/template.c
 */
int adsgm_init_obj_data(int chipnr, int objnr)
{
	chips_p[chipnr]->msgobj[objnr]->obj_base_addr=
      chips_p[chipnr]->chip_base_addr + (objnr + 1) * 0x10 ;
	chips_p[chipnr]->msgobj[objnr]->flags=0;
	
	return 0;
}

/**
 * adsgm_program_irq - program interrupts
 * @card: Number of the hardware card.
 *
 * The function adsgm_program_irq() is used for hardware that uses 
 * programmable interrupts. If your hardware doesn't use programmable interrupts
 * you should not set the @candevices_t->flags entry to %PROGRAMMABLE_IRQ and 
 * leave this function unedited. Again this function is hardware specific so 
 * there's no example code.
 * Return value: The function returns zero on success or %-ENODEV on failure
 * File: src/template.c
 */
int adsgm_program_irq(int card)
{
	return 0;
}

/**
 * adsgm_write_register - Low level write register routine
 * @data: data to be written
 * @address: memory address to write to
 *
 * The function adsgm_write_register() is used to write to hardware registers
 * on the CAN chip. You should only have to edit this function if your hardware
 * uses some specific write process.
 * Return Value: The function does not return a value
 * File: src/template.c
 */
void adsgm_write_register(unsigned char data, unsigned long addr)
{
  //unsigned int address = ( ( addr << 2 ) + base_io );
  unsigned long address = ( addr & 0xFFFF0000 ) + (( addr & 0x00003FFF) << 2 );
    
  if (address < base_io || address > base_io+IO_RANGE) {
    printk (KERN_ERR "w -- invalid addr %08lX\n",address);

    return;
  }
  
  outb(data, address );
  udelay(2);
}

/**
 * adsgm_read_register - Low level read register routine
 * @address: memory address to read from
 *
 * The function adsgm_read_register() is used to read from hardware registers
 * on the CAN chip. You should only have to edit this function if your hardware
 * uses some specific read process.
 * Return Value: The function returns the value stored in @address
 * File: src/template.c
 */
unsigned adsgm_read_register(unsigned long addr)
{
  unsigned long flags;
  unsigned char data;
  //unsigned int address = ( ( addr << 2 ) + base_io );
  unsigned long address = ( addr & 0xFFFF0000 ) + (( addr & 0x00003FFF) << 2 );

  save_flags_cli(flags);

  if (address < base_io || address > base_io+IO_RANGE) {
    printk (KERN_ERR "-- invalid addr %08lX\n",address);

    return 0;
  }

  /* se if we are reading from RAM */
  if (address >= ( base_io + (IO_RAM << 2) )) {
    /* first do a dummy read */
    data = inb( address );

    /* read from high speed register */
    data = inb( base_io + (iHSR << 2) );
  } else {
    data = inb( address );
  }

  restore_flags(flags);
  return data;
}

/* !!! Don't change this function !!! */
int adsgm_register(struct hwspecops_t *hwspecops)
{
	hwspecops->request_io = adsgm_request_io;
	hwspecops->release_io = adsgm_release_io;
	hwspecops->reset = adsgm_reset;
	hwspecops->init_hw_data = adsgm_init_hw_data;
	hwspecops->init_chip_data = adsgm_init_chip_data;
	hwspecops->init_obj_data = adsgm_init_obj_data;
	hwspecops->write_register = adsgm_write_register;
	hwspecops->read_register = adsgm_read_register;
	hwspecops->program_irq = adsgm_program_irq;
	return 0;
}
